#ifndef NFS_CLUSTER_H
#define NFS_CLUSTER_H


/*
 * Valid flags for a dirty buffer
 */
#define PG_UNLOCK_AFTER         0x0001
#define PG_CANCELLED            0x0002
#define PG_INVALIDATE_AFTER     0x0004

#ifdef __KERNEL__
struct nfs_page {
        struct page *           page;
        unsigned short          flags;

        struct nfs_cluster *    cluster;        /* Back pointer to cluster */
        struct rpc_task *       task;
        u64                     start;
        u64                     end;

        unsigned long           timeout;        /* when to write/commit */

        unsigned short          commit;         /* batch of commit pages */
};


typedef int (*nfs_flush_req)(struct nfs_page *, int);


/*
 * Counters of total number and pending number of requests.
 * When the total number of requests exceeds the soft limit, we start
 * flushing out requests. If it exceeds the hard limit, we stall until
 * it drops again.
 */
#define MAX_REQUEST_SOFT        128
#define MAX_REQUEST_HARD        1024

/*
 * Maximum number of requests per write cluster.
 * 32 requests per cluster account for 128K of data on an intel box.
 * Note: it's a good idea to make this number smaller than MAX_REQUEST_SOFT.
 *
 * For 100Mbps Ethernet, 128 pages (i.e. 256K) per cluster gives much
 * better performance.
 */
#define CLUSTER_SHIFT           7
#define CLUSTER_PAGES           (1 << CLUSTER_SHIFT)
#define CLUSTER_SIZE            (CLUSTER_PAGES * PAGE_SIZE)
#define CLUSTER_MASK            (~(CLUSTER_SIZE - 1))
#define CLUSTER_NR(off)         ((off) >> (CLUSTER_SHIFT + PAGE_SHIFT))
#define CLUSTER_HASH_SIZE       16
#define CLUSTER_HASH(ino, off)  (((ino) ^ CLUSTER_NR(off)) & (CLUSTER_HASH_SIZE - 1))
#define REQUEST_NR(off)         (((off) >> PAGE_SHIFT) & (CLUSTER_PAGES - 1))


/*
 * This is a cluster of read/write requests that should be sent to
 * the server together.
 *
 * Note that there may be more than one cluster for the same file
 * region, but with different credentials and pid.
 */
struct nfs_cluster {
        struct rpc_listitem     cl_list;     /* list of clusters per inode */
        struct nfs_cluster *    hash_prev;      /* hash chains */
        struct nfs_cluster *    hash_next;

        /* List of requests */
        struct nfs_page *       request[CLUSTER_PAGES];
	nfs_flush_req           flush;

        struct file *           file;           /* file */
        __u64                   start;          /* file range covered */
        __u64                   end;
        unsigned long           nextscan;       /* when to flush aged bufs */
        unsigned short          count;          /* # of references */
        unsigned short          pages;          /* # of requests total */
        unsigned short          dirty;          /* # of dirty requests */
        unsigned short          pending;        /* # of pending WRITE calls */
        struct wait_queue *     wait;
#define MAX_COMMIT_SEQ          0xFFFF
#define IS_DIRTY(req)           ((req)->commit == MAX_COMMIT_SEQ)	
};


#define CL_NEXT(c)	((struct nfs_cluster *)((c)->cl_list.next))

/*
 * Functions
 */
extern int               sync_cluster(struct nfs_cluster *, int);
extern int               scan_cluster(struct nfs_cluster *, unsigned int);
extern int               wait_cluster(struct nfs_cluster *);
extern void              delete_cluster(struct nfs_cluster *);

extern int               nfs_strategy(struct nfs_cluster *);

static inline void
release_cluster(struct nfs_cluster *cluster)
{
        if (cluster && --(cluster->count) == 0)
                delete_cluster(cluster);
}

/* Request functions */
extern struct nfs_page * nfs_create_request(struct nfs_cluster *,
					    struct page *,
					    __u64, __u64, int *);
extern void              nfs_release_request(struct nfs_page *);
extern int               nfs_update_request(struct file *,
					    struct page *,
					    __u64, __u64,
					    struct nfs_page **,
					    nfs_flush_req);
extern int               nfs_wait_on_request(struct nfs_page *);
extern void              nfs_cancel_request(struct nfs_cluster *,
					    struct nfs_page *);

/*
 * Mark a request dirty.
 */
static inline void
nfs_mark_request_dirty(struct nfs_cluster *cluster, struct nfs_page *req)
{
        mark_inode_dirty(cluster->file->f_dentry->d_inode);
        if (req->commit != MAX_COMMIT_SEQ)
                cluster->dirty++;
        req->commit = MAX_COMMIT_SEQ;
}

/*
 * Lock the page of an asynchronous request
 */
static inline int
nfs_lock_page(struct nfs_page *req)
{
        struct page *page = req->page;

        if (test_and_set_bit(PG_locked, &page->flags))
		return 0;
        req->flags |= PG_UNLOCK_AFTER;
        return 1;
}

static inline void
nfs_unlock_page(struct page *page)
{
	clear_bit(PG_locked, &page->flags);
	wake_up(&page->wait);
}
#endif



#endif
